/* 
 * HPCU Bi-Weekly Challenge: Hybrid Parallel Acceleration
 * Author: Aaron Weeden, Shodor, 2014
 * Last Modified: September 13, 2014, by Aaron Weeden
 */

/* import libraries */
#include <omp.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

/* create macro for exiting if MPI errors occur */
#define tryMPI(ret, str, rank) \
    if (ret != MPI_SUCCESS) { \
        fprintf(stderr, "Rank %d returned error for %s, exiting!\n", \
                rank, str); \
        exit(EXIT_FAILURE); \
    }

/* declare constants */
const double ACCELERATION = -9.81;
const double VELOCITY_0 = 0.0;
const double FACTOR = 1e6;
const int MPI_TAG = 0;

/* start program */
int main(int argc, char **argv) {

    /* declare variables */
    int mpi_rank;
    int mpi_size;
    int num_threads;
    double velocity;
    double position;
    int t;
    int i;
    int j;
    int *final_times;

    /* set up mpi */
    tryMPI(
            MPI_Init(&argc, &argv),
            "MPI_Init",
            0
          );
    tryMPI(
            MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank),
            "MPI_Comm_rank",
            0
          );
    tryMPI(
            MPI_Comm_size(MPI_COMM_WORLD, &mpi_size),
            "MPI_Comm_size",
            mpi_rank
          );

    /* get number of threads */
#pragma omp parallel
    {
#pragma omp single
        {
            num_threads = omp_get_num_threads();
        }
    }

    /* allocate memory for array (enough space for each thread) */
    if ((final_times = malloc(num_threads * sizeof(int))) == NULL) {
        fprintf(stderr, "Rank %d returned error for malloc, exiting!\n",
                mpi_rank);
        exit(EXIT_FAILURE);
    }

    /* initialize velocity */
    velocity = VELOCITY_0;

    /* spawn OpenMP threads */
#pragma omp parallel private(position, velocity, t)
    {
        /* initialize position */
        position = FACTOR * (mpi_rank * num_threads + omp_get_thread_num());

        /* run the simulation */
        while (position > 0) {
            velocity += ACCELERATION;
            position += velocity;
            t++;
        }

        /* store the final time in an array */
        final_times[omp_get_thread_num()] = t;

        /* done spawning threads */
    }

    /* send array to rank 0 */
    tryMPI(
            MPI_Send(final_times, num_threads, MPI_INT, 0, MPI_TAG,
                MPI_COMM_WORLD),
            "MPI_Send",
            mpi_rank
          );

    /* rank 0 only */
    if (mpi_rank == 0) {

        /* receive array from each process */
        for (i = 0; i < mpi_size; i++) {
            tryMPI(
                    MPI_Recv(final_times, num_threads, MPI_INT, i,
                        MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE),
                    "MPI_Recv",
                    mpi_rank
                  );

            /* print results */
            for (j = 0; j < num_threads; j++) {
                printf("%f %d\n",
                        FACTOR * (i*num_threads+j),
                        final_times[j]);
            }
        }
    }

    /* free memory for array */
    free(final_times);

    /* tear down mpi */
    tryMPI(
            MPI_Finalize(),
            "MPI_Finalize",
            mpi_rank
          );

    /* quit program */
    return 0;
}
